#define USE_ARDUINO_INTERRUPTS true    // Configurare intreruperi hardware pentru a monitoriza semnalul de la senzorul de puls
#include <PulseSensorPlayground.h>
#include <Wire.h>
#include <LiquidCrystal_I2C.h>
#include <SPI.h>
#include <SD.h>
#include <avr/wdt.h>                   // Include biblioteca Watchdog Timer

// Configurare senzor de puls
const int PulseWire = 0;               // Pinul de semnal 'S' al senzorului de puls conectat la A0
int Threshold = 550;                   // Determina care semnal sa fie "contorizat ca bataie" si care sa fie ignorat
PulseSensorPlayground pulseSensor;     // Creeaza un obiect

// Configurare LCD cu adresa 0x27
LiquidCrystal_I2C lcd(0x27, 16, 2);

#define MEASURING_PERIOD 20         // Constanta pentru perioada de timp la care se masoara EKG-ul
#define EKG_SENSOR A1               // Pinul de OUTPUT al senzorului EKG conectat la A1
#define LO_MINUS 3                  // Pinul LO- conectat la pinul digital 3
#define LO_PLUS 4                   // Pinul LO+ conectat la pinul digital 4

// Variabile pentru stocarea starii anterioare a electrozilor
int previousLoMinusState = LOW;
int previousLoPlusState = LOW;
int lastMeasureTime;

// Configurare card SD
const int chipSelect = 10;
File dataFile;

void setup() {
    Serial.begin(9600);

    // Dezactiveaza watchdog timer-ul daca a fost activat anterior
    wdt_disable();

    lcd.init();
    lcd.backlight();
    lcd.setCursor(0, 0);
    lcd.print("Pulse sensor is");
    lcd.setCursor(0, 1);
    lcd.print("initialized!");

    // Initializeaza cardul SD
    if (!SD.begin(chipSelect)) {
        lcd.clear();
        Serial.println("Initialization failed!");
        lcd.setCursor(0, 0);
        lcd.print("SD init failed!");
        delay(2000);
        resetSystem();
    }

    // Configureaza obiectul PulseSensor cu threshold si pinul de intrare analogic alese
    pulseSensor.analogInput(PulseWire);   
    pulseSensor.setThreshold(Threshold);   

    // Verific daca obiectul "pulseSensor" a fost creat
    if (pulseSensor.begin()) {
        Serial.println("PulseSensor object created!");
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print("Pulse sensor");
        lcd.setCursor(0, 1);
        lcd.print("is ready");
        delay(5000);
        lcd.clear();
    }

    // Seteaza pinii LO+ si LO- ca intrari pentru detectarea deconectarii electrozilor
    pinMode(LO_MINUS, INPUT);
    pinMode(LO_PLUS, INPUT);
}

void loop() {
    // Numaratoare inversa 2 minute
    countdown(2);

    // Masoara valorile pulsului timp de 1 minut
    measurePulse(1);

    // Numaratoare inversa 1 minut
    countdown(1);

    // Masoara valorile EKG-ului
    measureEKG();

    // Reseteaza sistemul
    resetSystem();
}

void countdown(int minutes) {
    for (int i = minutes * 60; i > 0; i--) {
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print("Start measuring");
        lcd.setCursor(0, 1);
        lcd.print("in ");
        lcd.print(i / 60);
        lcd.print("m ");
        lcd.print(i % 60);
        lcd.print("s");
        delay(1000);
    }
}

void measurePulse(int minutes) {
    unsigned long startTime = millis();
    unsigned long endTime = startTime + minutes * 60000;

    int sumBPM = 0;
    int countBPM = 0;

    lcd.clear();
    lcd.setCursor(3, 0);
    lcd.print("Measuring");
    lcd.setCursor(4, 1);
    lcd.print("pulse...");

    while (millis() < endTime) {
        // Calculez BPM
        int myBPM = pulseSensor.getBeatsPerMinute();

        // Testez daca a avut loc o bataie
        if (pulseSensor.sawStartOfBeat()) {
            // Afiseaz valoarea BPM
            Serial.println("♥  A HeartBeat Happened ! ");
            Serial.print("BPM: ");
            Serial.println(myBPM);

            // Salveaza valorile BPM relevante pe cardul SD
            if (myBPM > 0) {
                writeData("Pulse", myBPM, "pulse.txt");

                sumBPM += myBPM;
                countBPM++;
            }
        }

        delay(20);
    }

    // Calculeaza media valorilor BPM
    float averageBPM = (float)sumBPM / countBPM;

    // Afiseaza media valorilor BPM pe LCD
    lcd.clear();
    lcd.setCursor(2, 0);
    lcd.print("Average BPM:");
    lcd.setCursor(5, 1);
    lcd.print(averageBPM);

    Serial.print("Average BPM: ");
    Serial.println(averageBPM);

    delay(10000);
}

void measureEKG() {
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("Measuring EKG");

    unsigned long startTime = millis();
    unsigned long currentTime = startTime;

    // Fac masuratori pentru EKG timp de 30 de secunde
    while (currentTime - startTime < 30000) {
        int loMinusState = digitalRead(LO_MINUS);
        int loPlusState = digitalRead(LO_PLUS);

        if (loMinusState != previousLoMinusState || loPlusState != previousLoPlusState) {
          // Actualizez starea anterioara a electrozilor
          previousLoMinusState = loMinusState;
          previousLoPlusState = loPlusState;

          // Verific daca unul dintre electrozi este deconectat
          if (loMinusState == HIGH || loPlusState == HIGH) {
            Serial.println("Electrozii nu sunt conectati!");
          }
        }

        // Daca a trecut perioada de asteptare intre masuratori
        if (millis() - lastMeasureTime > MEASURING_PERIOD) {
          if (loMinusState == LOW && loPlusState == LOW) {
            // Citeste valoarea EKG
            int ekgValue = analogRead(EKG_SENSOR);
            Serial.println(ekgValue);
            // writeData("EKG", ekgValue, "ekg.txt");
          }

          lastMeasureTime = millis();
        }

        delay(10);
        currentTime = millis();
    }
}

// Scriere date pe cardul SD
void writeData(String dataType, int value, String file) {
    dataFile = SD.open(file, FILE_WRITE);

    if (dataFile) {
        dataFile.print(dataType);
        dataFile.print(": ");
        dataFile.println(value);
        dataFile.close();
        Serial.println("Data written to SD card.");
    } else {
        Serial.println("Error opening data file.");
    }
}

// Resetare sistem
void resetSystem() {
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("Resetting...");

    // Activeaza watchdog timer-ul cu un timeout de 1 secunda
    wdt_enable(WDTO_1S);
    while (1); // Asteapta ca watchdog timer-ul sa reseteze sistemul
}